/**********************************************************************
     xbm2cif.c:  Simple X-Bitmap to cif converter

     Purpose:  For drawing bitmaps in metal1 on VLSI circuit layouts
     Author:   Tim Edwards
     Date:  Sometime around 1991.
     Updated 2015 just because my 1991 programming skills were so horrid.

***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

main(int argc, char *argv[])
{
    FILE *fin, *fout;

    char buf, line_in[256], *sptr;
    char use_stdin, use_stdout;
    int width, height, i, j, w, boxx, boxy;
    int totbytes;
    unsigned short *bits, mask;
    int newwidth, size;
    float scale;
    
    scale = 1.0;

    if (argc > 1) {
       if (sscanf(argv[1], "%f", &scale) == 0) {
          if (!strncmp(argv[1], "-h", 2) || !strncmp(argv[1], "--h", 3)) {
	     fprintf(stderr, "Usage:  xbmtocif [scale] file.xbm file.cif\n");
	     exit(0);
          }
	  fprintf(stderr, "Warning:  bad scale, using 1.0\n");
	  scale = 1.0;
       }
    }
    if (argc > 2) {
       fin = fopen(argv[2], "r");
       if (fin == NULL) {
	  fprintf(stderr, "Error:  Cannot open file \"%s\" for reading.\n",
		argv[2]);
	  exit(1);
       }
    }
    else fin = stdin;

    if (argc > 3) {
       fout = fopen(argv[3], "w");
       if (fout == NULL) {
	  fprintf(stderr, "Error:  Cannot open file \"%s\" for writing.\n",
		argv[3]);
	  exit(1);
       }
    }
    else fout = stdout;

    /* read xbm */

    if (fscanf(fin, "%*s %*s %d", &width) != EOF) 
	fscanf(fin, "%*s %*s %d", &height);
    else {
	fprintf(stderr, "Error:  Input is not an xbm file?\n");
	exit(1);
    }

    newwidth = (int)((width + 7) / 8.0) * 8;
    mask = (1 << (newwidth - width)) - 1;
    totbytes = (newwidth * height) / 8;
    fprintf(fout, "(Bitmap width = %d;  Bitmap height = %d);\n",
		newwidth, height);
    fprintf(fout, "(Total input bytes = %d);\n", totbytes);
    fprintf(fout, "(Scale = %6.4f);\n", scale);
    
    if ((bits = (unsigned short *)malloc(totbytes
			* sizeof(unsigned short))) == NULL) {
	fprintf(stderr, "Error: Out of memory\n");
    }

    while ((buf = getc(fin)) != '{');
    w = 0;
    for (i = 0; i < totbytes;) { 
	if (fscanf(fin, "%s", line_in) == EOF) break;
	sptr = line_in;
	while (sscanf(sptr, "0x%hx", &bits[i]) == 1) {
	   if (width != newwidth) {
	      w += 8;
	      if (w > width) {
		 bits[i] &= mask;
		 w = 0;
	      }
	   }
	   sptr = strchr(sptr, ',');
	   if (sptr == NULL) break;
	   sptr++;
	   i++;
	   if (i >= totbytes) break;
	}
    }
    if (fin != stdin) fclose(fin);
   
    /* write cif */

    /* write layer (and scale factor) */

    fprintf(fout, "DD 300;\n DS 300 %d 1;\n", (int)(scale * 50));
    fprintf(fout, "9 bitmap;\n L CMF;\n"); 
    size = 1;
    for (i = 0; i < totbytes; i++) {
	for (j = 0; j < 8; j++) {
	    if ((bits[i] & 0x01) != 0) {
		boxy = -2 * (int)(((i * 8) + j) / newwidth);
		boxx = 2 * (((i * 8) + j) % newwidth);
		if ((bits[i] & 0x02) != 0)
		    size++;
		else if ((j == 7) && ((bits[i + 1] & 0x01) != 0)
				&& (boxx < newwidth - 1))
		    size++;
		else {
		    fprintf(fout, "B %d %d %d %d;\n",
				2 * size, 2, boxx - size, boxy);
		    size = 1;
		}
	    }
	    bits[i] >>= 1;
	}
    }
    fprintf(fout, "DF;\n C 300;\n");
    fprintf(fout, "End\n");

    if (fout != stdout) fclose(fout);
    free(bits);
    exit(0);
}
